@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE BangPatterns #-}
 {-# LANGUAGE GADTs #-}
 {-# LANGUAGE RankNTypes #-}
@@ -26,13 +27,19 @@ module Regex.Internal.Parser
   , parseNext
   ) where
 
-import Control.Applicative
+import Control.Applicative ((<|>), empty)
+import qualified Control.Applicative as Ap
 import Control.Monad.Trans.State.Strict
-import Control.Monad.Fix
+  ( State, StateT, evalState, evalStateT, execState, gets, modify', state)
+import Control.Monad.Fix (mfix)
 import Data.Maybe (isJust)
-import Data.Primitive.SmallArray
 import qualified Data.Foldable as F
+import qualified Data.Traversable as T
+#ifdef __GLASGOW_HASKELL__
+import Data.Primitive.SmallArray
+  (SmallArray, emptySmallArray, smallArrayFromList)
 import qualified GHC.Exts as X
+#endif
 
 import Regex.Internal.Regex (RE(..), Strictness(..), Greediness(..))
 import Regex.Internal.Unique (Unique(..), UniqueSet)
@@ -88,14 +95,14 @@ compileToParser re = case re of
   RFmap_ a re1 -> PFmap_ <$> compileToNode a re1
   RPure a -> pure $ PPure a
   RLiftA2 st f re1 re2 ->
-    liftA2 (PLiftA2 st f) (compileToParser re1) (compileToParser re2)
+    Ap.liftA2 (PLiftA2 st f) (compileToParser re1) (compileToParser re2)
   REmpty -> pure PEmpty
   RAlt re01 re02 -> do
     u <- nxtU
     let (re1,re2,res) = gatherAlts re01 re02
     p1 <- compileToParser re1
     p2 <- compileToParser re2
-    ps <- traverse compileToParser res
+    ps <- T.traverse compileToParser res
     pure $ PAlt u p1 p2 (smallArrayFromList ps)
   RFold st gr f z re1 -> do
     u <- nxtU
@@ -125,7 +132,7 @@ compileToNode a re0 = go re0 (NAccept a)
             (re1,re2,res) = gatherAlts re01 re02
         n1 <- go re1 nxt1
         n2 <- go re2 nxt1
-        ns <- traverse (flip go nxt1) res
+        ns <- T.traverse (flip go nxt1) res
         pure $ NAlt n1 n2 (smallArrayFromList ns)
       RFold _ gr _ _ re1 -> goMany gr re1 nxt
       RMany _ _ _ _ re1 -> goMany Greedy re1 nxt
@@ -142,10 +149,10 @@ compileToNode a re0 = go re0 (NAccept a)
 gatherAlts :: RE c a -> RE c a -> (RE c a, RE c a, [RE c a])
 gatherAlts re01 re02 = case go re01 (go re02 []) of
   re11:re12:res -> (re11, re12, res)
-  _ -> errorWithoutStackTrace "Regex.Internal.Parser.gatherAlts: impossible"
+  _ -> error "Regex.Internal.Parser.gatherAlts: impossible"
   where
-    go (RAlt re1 re2) = go re1 . go re2
-    go re = (re:)
+    go (RAlt re1 re2) acc = go re1 (go re2 acc)
+    go re acc = re:acc
 
 --------------------
 -- Compile bounded
@@ -182,10 +189,10 @@ checkSize lim re0 = isJust (evalStateT (go re0) 0)
         RMany _ _ _ _ re1 -> inc *> go re1
         RFold _ _ _ _ re1 -> inc *> go re1
     inc = do
-      n <- get
-      if n == lim
-      then empty
-      else put $! n+1
+      ok <- gets (< lim)
+      if ok
+      then modify' (+1)
+      else empty
 
 ----------
 -- Parse
@@ -263,17 +270,19 @@ down p !ct !pt = case p of
         modify' $ up x ct
 
 downNode :: Node c b -> Cont c b a -> StepState c a -> StepState c a
-downNode n0 !ct = go n0
-  where
-    go n !pt = case n of
-      NAccept b -> up b ct pt
-      NGuard u n1
-        | U.member u (sSet pt) -> pt
-        | otherwise -> go n1 (pt { sSet = U.insert u (sSet pt) })
-      NToken t nxt ->
-        pt { sNeed = NeedCCons t (CFmap_ nxt ct) (sNeed pt) }
-      NEmpty -> pt
-      NAlt n1 n2 ns -> F.foldl' (flip go) (go n2 (go n1 pt)) ns
+downNode n !ct !pt = case n of
+  NAccept b -> up b ct pt
+  NGuard u n1
+    | U.member u (sSet pt) -> pt
+    | otherwise -> downNode n1 ct (pt { sSet = U.insert u (sSet pt) })
+  NToken t nxt ->
+    pt { sNeed = NeedCCons t (CFmap_ nxt ct) (sNeed pt) }
+  NEmpty -> pt
+  NAlt n1 n2 ns ->
+    F.foldl'
+      (\pt' n' -> downNode n' ct pt')
+      (downNode n2 ct (downNode n1 ct pt))
+      ns
 
 up :: b -> Cont c b a -> StepState c a -> StepState c a
 up b ct !pt = case ct of
@@ -356,14 +365,14 @@ prepareParser p = toParserState (down p CTop stepStateZero)
 -- Returns @Nothing@ if parsing has failed regardless of further input.
 -- Otherwise, returns an updated @ParserState@.
 stepParser :: ParserState c a -> c -> Maybe (ParserState c a)
-stepParser ps c = case psNeed ps of
+stepParser ps c0 = case psNeed ps of
   NeedCNil -> Nothing
-  needs -> toParserState (go needs)
+  needs -> toParserState (go c0 needs)
   where
-    go (NeedCCons t ct rest) =
-      let !pt = go rest
+    go c (NeedCCons t ct rest) =
+      let !pt = go c rest
       in maybe pt (\b -> up b ct pt) (t c)
-    go NeedCNil = stepStateZero
+    go _ NeedCNil = stepStateZero
 {-# INLINE stepParser #-}
 
 -- | \(O(1)\). Get the parse result for the input fed into the parser so far.
@@ -409,7 +418,11 @@ type Foldr f a = forall b. (a -> b -> b) -> b -> f -> b
 parseFoldr :: Foldr f c -> Parser c a -> f -> Maybe a
 parseFoldr fr = \p xs -> prepareParser p >>= fr f finishParser xs
   where
-    f c k = X.oneShot (\ !ps -> stepParser ps c >>= k)
+    f c k =
+#ifdef __GLASGOW_HASKELL__
+      X.oneShot
+#endif
+        (\ !ps -> stepParser ps c >>= k)
 {-# INLINE parseFoldr #-}
 
 -- | \(O(mn \log m)\). Run a parser given a \"@next@\" action.
@@ -476,6 +489,20 @@ unlessM mb mx = do
   b <- mb
   if b then pure () else mx
 
+-----------------
+-- Array compat
+-----------------
+
+#ifndef __GLASGOW_HASKELL__
+type SmallArray = []
+
+emptySmallArray :: SmallArray a
+emptySmallArray = []
+
+smallArrayFromList :: [a] -> SmallArray a
+smallArrayFromList = id
+#endif
+
 ----------
 -- Notes
 ----------
