@@ -44,12 +44,14 @@ module Regex.Internal.Regex
   , foldlManyMin'
   ) where
 
-import Control.Applicative
+import Control.Applicative (Alternative(..))
+import qualified Control.Applicative as Ap
 import Control.DeepSeq (NFData(..), NFData1(..), rnf1)
-import Control.Monad
+import Control.Monad (void)
 import Data.Functor.Classes (Eq1(..), Ord1(..), Show1(..), showsUnaryWith)
 import Data.Semigroup (Semigroup(..))
 import qualified Data.Foldable as F
+import qualified Data.Traversable as T
 
 ---------------------------------
 -- RE and constructor functions
@@ -95,7 +97,7 @@ data RE c a where
   RPure   :: a -> RE c a
   RLiftA2 :: !Strictness -> !(a1 -> a2 -> a) -> !(RE c a1) -> !(RE c a2) -> RE c a
   REmpty  :: RE c a
-  RAlt    :: !(RE c a) -> !(RE c a) -> (RE c a)
+  RAlt    :: !(RE c a) -> !(RE c a) -> RE c a
   RFold   :: !Strictness -> !Greediness -> !(a -> a1 -> a) -> a -> !(RE c a1) -> RE c a
   RMany   :: !(a1 -> a) -> !(a2 -> a) -> !(a2 -> a1 -> a2) -> !a2 -> !(RE c a1) -> RE c a -- Strict and greedy implicitly
 
@@ -112,8 +114,8 @@ fmap' = RFmap Strict
 instance Applicative (RE c) where
   pure = RPure
   liftA2 = RLiftA2 NonStrict
-  re1 *> re2 = liftA2 (const id) (void re1) re2
-  re1 <* re2 = liftA2 const re1 (void re2)
+  re1 *> re2 = Ap.liftA2 (const id) (void re1) re2
+  re1 <* re2 = Ap.liftA2 const re1 (void re2)
 
 liftA2' :: (a1 -> a2 -> b) -> RE c a1 -> RE c a2 -> RE c b
 liftA2' = RLiftA2 Strict
@@ -126,14 +128,14 @@ instance Alternative (RE c) where
 
 -- | @(<>)@ = @liftA2 (<>)@
 instance Semigroup a => Semigroup (RE c a) where
-  (<>) = liftA2 (<>)
-  sconcat = fmap sconcat . sequenceA
+  (<>) = Ap.liftA2 (<>)
+  sconcat = fmap sconcat . T.sequenceA
   {-# INLINE sconcat #-}
 
 -- | @mempty@ = @pure mempty@
 instance Monoid a => Monoid (RE c a) where
   mempty = pure mempty
-  mconcat = fmap mconcat . sequenceA
+  mconcat = fmap mconcat . T.sequenceA
   {-# INLINE mconcat #-}
 -- Use the underlying type's sconcat/mconcat because it may be more efficient
 -- than the default right-associative definition.
@@ -216,7 +218,7 @@ instance Functor Many where
     Repeat x -> Repeat (f x)
     Finite xs -> Finite (map f xs)
 
-instance Foldable Many where
+instance F.Foldable Many where
   foldr f z m = case m of
     Repeat x -> let r = f x r in r
     Finite xs -> foldr f z xs
@@ -331,23 +333,23 @@ sepEndBy re sep = sepEndBy1 re sep <|> pure []
 -- | @r \`sepEndBy1\` sep@ parses one or more occurences of @r@, separated and
 -- optionally ended by @sep@. Biased towards matching more.
 sepEndBy1 :: RE c a -> RE c sep -> RE c [a]
-sepEndBy1 re sep = sepBy1 re sep <* optional sep
+sepEndBy1 re sep = sepBy1 re sep <* Ap.optional sep
 
 -- | @chainl1 r op@ parses one or more occurences of @r@, separated by @op@.
 -- The result is obtained by left associative application of all functions
 -- returned by @op@ to the values returned by @p@. Biased towards matching more.
 chainl1 :: RE c a -> RE c (a -> a -> a) -> RE c a
-chainl1 re op = liftA2 (flip id) re rest
+chainl1 re op = Ap.liftA2 (flip id) re rest
   where
-    rest = foldlMany (flip (.)) id (liftA2 flip op re)
+    rest = foldlMany (flip (.)) id (Ap.liftA2 flip op re)
 
 -- | @chainr1 r op@ parses one or more occurences of @r@, separated by @op@.
 -- The result is obtained by right associative application of all functions
 -- returned by @op@ to the values returned by @p@. Biased towards matching more.
 chainr1 :: RE c a -> RE c (a -> a -> a) -> RE c a
-chainr1 re op = liftA2 id rest re
+chainr1 re op = Ap.liftA2 id rest re
   where
-    rest = foldlMany (.) id (liftA2 (flip id) re op)
+    rest = foldlMany (.) id (Ap.liftA2 (flip id) re op)
 
 -- | Results in the first occurence of the given @RE@. Fails if no occurence
 -- is found.
